package com.smppw.core;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.collection.ListUtil;
import cn.hutool.core.map.MapUtil;
import com.smppw.common.pojo.IndexesTradeDate;
import com.smppw.common.pojo.dto.CompoundRet;
import com.smppw.common.pojo.dto.DateValue;
import com.smppw.common.pojo.dto.calc.IndicatorCalcPropertyDto;
import com.smppw.common.pojo.dto.calc.IndicatorCalcReqPropertyDto;
import com.smppw.common.pojo.dto.calc.IndicatorCalcTimeRangeDto;
import com.smppw.common.pojo.dto.indicator.CalcMultipleSecMultipleTimeRangeIndicatorReq;
import com.smppw.common.pojo.dto.indicator.DateIntervalDto;
import com.smppw.common.pojo.enums.*;
import com.smppw.common.pojo.enums.strategy.IStrategy;
import com.smppw.common.pojo.enums.strategy.Strategy;
import com.smppw.utils.DateIntervalUtil;
import com.smppw.utils.IndicatorCalcUtil;
import com.smppw.utils.NavUtil;
import com.smppw.utils.TradeDateUtil;
import org.apache.commons.lang3.StringUtils;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * @author wangzaijun
 * @date 2023/8/2 10:17
 * @description 指标计算和走势图
 */
public class IndicatorService {
    private IndicatorService() {
    }

    public static IndicatorService getInstance() {
        return IndicatorService.IndicatorServiceBuilder.INSTANCE;
    }

    public Map<String, List<IndicatorCalcPropertyDto>> getMultipleSecTrend(List<String> mainSecIdList, Map<String, String> secBenchmarkIdMap, List<String> indexIdList,
                                                                           DateIntervalDto dateIntervalDto, Frequency frequency, List<TrendType> trendTypeV2List,
                                                                           Map<String, Frequency> secFreqMap, Map<String, List<DateValue>> allNavMap) {
        return this.getMultipleSecTrend(mainSecIdList, secBenchmarkIdMap, indexIdList, dateIntervalDto, null, null, frequency, trendTypeV2List, secFreqMap, allNavMap);
    }

    /**
     * 计算多标的的业绩走势图
     *
     * @param mainSecIdList
     * @param secBenchmarkIdMap
     * @param indexIdList
     * @param dateIntervalDto
     * @param fixedIncome
     * @param initValue
     * @param frequency
     * @param trendTypeV2List
     * @param secFreqMap
     * @param allNavMap
     * @return
     */
    public Map<String, List<IndicatorCalcPropertyDto>> getMultipleSecTrend(List<String> mainSecIdList, Map<String, String> secBenchmarkIdMap, List<String> indexIdList,
                                                                           DateIntervalDto dateIntervalDto, BigDecimal fixedIncome, BigDecimal initValue, Frequency frequency,
                                                                           List<TrendType> trendTypeV2List, Map<String, Frequency> secFreqMap, Map<String, List<DateValue>> allNavMap) {
        // 实际频率时净值还是取日频的
        Frequency navFrequency = frequency == null || frequency == Frequency.Default ? Frequency.Daily : frequency;
        Map<String, List<IndicatorCalcPropertyDto>> indicatorCalcPropertyDtoListMap = MapUtil.newHashMap(mainSecIdList.size());
        for (String mainSecId : mainSecIdList) {
            String benchmarkId = secBenchmarkIdMap.get(mainSecId);

            Map<String, List<DateValue>> secAllNavMap = MapUtil.newHashMap(indexIdList.size() + 3);
            secAllNavMap.put(mainSecId, NavUtil.getNewNavList(allNavMap.get(mainSecId)));
            secAllNavMap.put(benchmarkId, NavUtil.getNewNavList(allNavMap.get(benchmarkId)));
            for (String indexId : indexIdList) {
                secAllNavMap.put(indexId, NavUtil.getNewNavList(allNavMap.get(indexId)));
            }

            List<DateValue> mainSecIdNavList = secAllNavMap.get(mainSecId);

            String theFirstNavDate = "";
            String theLastNavDate = "";
            if (mainSecIdNavList != null && mainSecIdNavList.size() > 0) {
                theFirstNavDate = mainSecIdNavList.get(0).getDate();
                theLastNavDate = mainSecIdNavList.get(mainSecIdNavList.size() - 1).getDate();
            }

            List<IndexesTradeDate> indexesTradeDateList = TradeDateUtil.listInfoByDateInterval(theFirstNavDate, theLastNavDate);

            Frequency defaultFrequency = secFreqMap.getOrDefault(mainSecId, Frequency.Daily);
            Frequency calcFrequency = frequency == null || frequency == Frequency.Default ? defaultFrequency : frequency;

            List<DateValue> validMainSecNavList = NavUtil.filterFrequencyNav(mainSecIdNavList, indexesTradeDateList, navFrequency);
            secAllNavMap.put(mainSecId, validMainSecNavList);

            List<IndicatorCalcTimeRangeDto> indicatorCalcTimeRangeDtoList = DateIntervalUtil.getIndicatorCalcTimeRangeDto(dateIntervalDto, validMainSecNavList, indexesTradeDateList
                    , dateIntervalDto.getId());

            List<IndicatorCalcPropertyDto> indicatorCalcPropertyDtoList = ListUtil.list(false);

            for (IndicatorCalcTimeRangeDto indicatorCalcTimeRangeDto : indicatorCalcTimeRangeDtoList) {

                Map<String, List<DateValue>> dateSecAllNavMap = MapUtil.newHashMap(indexIdList.size() + 3);
                dateSecAllNavMap.put(mainSecId, NavUtil.getNewNavList(secAllNavMap.get(mainSecId)));
                dateSecAllNavMap.put(benchmarkId, NavUtil.getNewNavList(secAllNavMap.get(benchmarkId)));
                for (String indexId : indexIdList) {
                    dateSecAllNavMap.put(indexId, NavUtil.getNewNavList(secAllNavMap.get(indexId)));
                }

                IndicatorCalcReqPropertyDto indicatorCalcReq = prepareTrendReq(mainSecId, benchmarkId, indexIdList, indicatorCalcTimeRangeDto,
                        dateSecAllNavMap, calcFrequency, defaultFrequency, trendTypeV2List, fixedIncome, initValue);
                if (validMainSecNavList.size() > 0) {
                    IndicatorCalcPropertyDto trend = CalcService.getInstance().getTrend(indicatorCalcReq);
                    indicatorCalcPropertyDtoList.add(trend);
                } else {
                    IndicatorCalcPropertyDto indicatorCalcPropertyDto = new IndicatorCalcPropertyDto();
                    indicatorCalcPropertyDto.setIndicatorCalcReq(indicatorCalcReq);
                    indicatorCalcPropertyDtoList.add(indicatorCalcPropertyDto);
                }
            }
            indicatorCalcPropertyDtoListMap.put(mainSecId, indicatorCalcPropertyDtoList);
        }
        return indicatorCalcPropertyDtoListMap;
    }

    /**
     * 多标的多时段多指标计算
     *
     * @param req        请求参数
     * @param allNavMap  所有标的的净值
     * @param secFreqMap 标的的净值频率
     * @return /
     */
    public Map<String, List<IndicatorCalcPropertyDto>> calcMultipleSecMultipleTimeRangeIndicator(CalcMultipleSecMultipleTimeRangeIndicatorReq req,
                                                                                                 Map<String, List<DateValue>> allNavMap,
                                                                                                 Map<String, Frequency> secFreqMap) {

        List<String> mainSecIdList = req.getMainSecIdList();
        Map<String, String> secBenchmarkIdMap = req.getSecBenchmarkIdMap();
        List<String> indexIdList = req.getIndexIdList();
        Map<String, List<DateIntervalDto>> secDateIntervalDtoListMap = req.getSecDateIntervalDtoListMap();
        Frequency dataFrequency = req.getDataFrequency();
        List<Indicator> indicatorList = req.getIndicatorList();
        List<Indicator> geoExtraindicatorList = req.getGeoExtraindicatorList();
        String riskOfFreeId = req.getRiskOfFreeId();
        Double riskOfFreeValue = req.getRiskOfFreeValue();
        boolean ifAnnualize = req.isIfAnnualize();
        boolean calcIndexRetIndicatorValue = req.isCalcIndexRetIndicatorValue();
        boolean needRet = req.isNeedRet();
        boolean needExtraRet = req.isNeedExtraRet();
        boolean ifConvertPerformanceConsistencyWord = req.isIfConvertPerformanceConsistencyWord();
        Map<String, List<IndicatorCalcPropertyDto>> indicatorCalcPropertyDtoListMap = MapUtil.newHashMap(mainSecIdList.size());

        Frequency navFrequency = dataFrequency == null || dataFrequency == Frequency.Default ? Frequency.Daily : dataFrequency;

        for (String mainSecId : mainSecIdList) {

            String benchmarkId = secBenchmarkIdMap.get(mainSecId);
            Map<String, List<DateValue>> secAllNavMap = MapUtil.newHashMap(indexIdList.size() + 3);
            secAllNavMap.put(mainSecId, NavUtil.getNewNavList(allNavMap.get(mainSecId)));
            secAllNavMap.put(benchmarkId, NavUtil.getNewNavList(allNavMap.get(benchmarkId)));
            for (String indexId : indexIdList) {
                secAllNavMap.put(indexId, NavUtil.getNewNavList(allNavMap.get(indexId)));
            }
            if (StringUtils.isNotEmpty(riskOfFreeId)) {
                secAllNavMap.put(riskOfFreeId, NavUtil.getNewNavList(allNavMap.get(riskOfFreeId)));
            }

            List<IndicatorCalcPropertyDto> indicatorCalcPropertyDtoList = ListUtil.list(false);

            List<DateValue> mainSecIdNavList = secAllNavMap.get(mainSecId);

            String theFirstNavDate = "";
            String theLastNavDate = "";
            if (mainSecIdNavList != null && mainSecIdNavList.size() > 0) {
                theFirstNavDate = mainSecIdNavList.get(0).getDate();
                theLastNavDate = mainSecIdNavList.get(mainSecIdNavList.size() - 1).getDate();
            }
            List<IndexesTradeDate> indexesTradeDateList = TradeDateUtil.listInfoByDateInterval(theFirstNavDate, theLastNavDate);

            List<DateValue> validMainSecNavList = NavUtil.filterFrequencyNav(mainSecIdNavList, indexesTradeDateList, navFrequency);
            secAllNavMap.put(mainSecId, validMainSecNavList);

            Frequency defaultFrequency = secFreqMap.getOrDefault(mainSecId, Frequency.Daily);
            Frequency calcFrequency = dataFrequency == null || dataFrequency == Frequency.Default ? defaultFrequency : dataFrequency;

            List<DateIntervalDto> dateIntervalDtoList = ListUtil.list(false);
            if (secDateIntervalDtoListMap != null && secDateIntervalDtoListMap.get(mainSecId) != null) {
                dateIntervalDtoList = secDateIntervalDtoListMap.get(mainSecId);
            }

            for (DateIntervalDto dateIntervalDto : dateIntervalDtoList) {

                rollFindStartDate(mainSecId, dateIntervalDto, indexesTradeDateList, allNavMap, calcFrequency);

                List<IndicatorCalcTimeRangeDto> indicatorCalcTimeRangeDtoList = DateIntervalUtil.getIndicatorCalcTimeRangeDto(dateIntervalDto, validMainSecNavList, indexesTradeDateList
                        , dateIntervalDto.getId());

                for (IndicatorCalcTimeRangeDto indicatorCalcTimeRangeDto : indicatorCalcTimeRangeDtoList) {

                    Map<String, List<DateValue>> dateSecAllNavMap = MapUtil.newHashMap(indexIdList.size() + 3);
                    dateSecAllNavMap.put(mainSecId, NavUtil.getNewNavList(secAllNavMap.get(mainSecId)));
                    dateSecAllNavMap.put(benchmarkId, NavUtil.getNewNavList(secAllNavMap.get(benchmarkId)));
                    for (String indexId : indexIdList) {
                        dateSecAllNavMap.put(indexId, NavUtil.getNewNavList(secAllNavMap.get(indexId)));
                    }
                    if (StringUtils.isNotEmpty(riskOfFreeId)) {
                        dateSecAllNavMap.put(riskOfFreeId, NavUtil.getNewNavList(secAllNavMap.get(riskOfFreeId)));
                    }

                    IndicatorCalcReqPropertyDto indicatorCalcReq = prepareIndicatorCalcReq(mainSecId, benchmarkId, indexIdList, indicatorCalcTimeRangeDto,
                            dateSecAllNavMap, riskOfFreeId, riskOfFreeValue, null, calcFrequency, defaultFrequency, indicatorList, geoExtraindicatorList, ifAnnualize, calcIndexRetIndicatorValue,
                            needRet, needExtraRet, ifConvertPerformanceConsistencyWord);

                    IndicatorCalcPropertyDto indicatorCalcPropertyDto = CalcService.getInstance().indicatorCalc(indicatorCalcReq);
                    indicatorCalcPropertyDtoList.add(indicatorCalcPropertyDto);

                }
            }

            indicatorCalcPropertyDtoListMap.put(mainSecId, indicatorCalcPropertyDtoList);
        }


        return indicatorCalcPropertyDtoListMap;
    }

    /**
     * 提供给指标计算服务的接口，支持多标的多频率的指标计算
     *
     * @param mainSecIdList
     * @param secBenchmarkIdMap
     * @param indexIdList
     * @param dateIntervalDtoList
     * @param dataFrequencyList
     * @param dataFrequencyIndicatorListMap
     * @param dataFrequencyGeoExtraIndicatorListMap
     * @param riskOfFreeId
     * @param riskOfFreeValue
     * @param ifAnnualize
     * @param calcIndexRetIndicatorValue
     * @param needRet
     * @param needExtraRet
     * @param leastRetNum
     * @param allNavMap
     * @param secFreqMap
     * @return
     */
    public Map<String, Map<Frequency, List<IndicatorCalcPropertyDto>>> calcMultipleSecMultipleTimeRangeIndicator(List<String> mainSecIdList, Map<String, String> secBenchmarkIdMap, List<String> indexIdList, List<DateIntervalDto> dateIntervalDtoList,
                                                                                                                 List<Frequency> dataFrequencyList, Map<Frequency, List<Indicator>> dataFrequencyIndicatorListMap,
                                                                                                                 Map<Frequency, List<Indicator>> dataFrequencyGeoExtraIndicatorListMap, String riskOfFreeId, Double riskOfFreeValue,
                                                                                                                 boolean ifAnnualize, boolean calcIndexRetIndicatorValue, boolean needRet, boolean needExtraRet,
                                                                                                                 int leastRetNum, Map<String, List<DateValue>> allNavMap, Map<String, Frequency> secFreqMap) {

        Map<String, Map<Frequency, List<IndicatorCalcPropertyDto>>> indicatorCalcPropertyDtoListMapMap = MapUtil.newHashMap(mainSecIdList.size());
        for (String mainSecId : mainSecIdList) {

            String benchmarkId = secBenchmarkIdMap.get(mainSecId);

            Map<Frequency, List<IndicatorCalcPropertyDto>> indicatorCalcPropertyDtoListMap = MapUtil.newHashMap(dataFrequencyList.size());
            for (Frequency dataFrequency : dataFrequencyList) {
                Map<String, List<DateValue>> secAllNavMap = MapUtil.newHashMap(3 + indexIdList.size());
                secAllNavMap.put(mainSecId, NavUtil.getNewNavList(allNavMap.get(mainSecId)));
                secAllNavMap.put(benchmarkId, NavUtil.getNewNavList(allNavMap.get(benchmarkId)));
                for (String indexId : indexIdList) {
                    secAllNavMap.put(indexId, NavUtil.getNewNavList(allNavMap.get(indexId)));
                }
                if (StringUtils.isNotEmpty(riskOfFreeId)) {
                    secAllNavMap.put(riskOfFreeId, NavUtil.getNewNavList(allNavMap.get(riskOfFreeId)));
                }

                Frequency navFrequency = dataFrequency == null || dataFrequency == Frequency.Default ? Frequency.Daily : dataFrequency;
                List<DateValue> mainSecIdNavList = secAllNavMap.get(mainSecId);

                String theFirstNavDate = "";
                String theLastNavDate = "";
                if (mainSecIdNavList != null && mainSecIdNavList.size() > 0) {
                    theFirstNavDate = mainSecIdNavList.get(0).getDate();
                    theLastNavDate = mainSecIdNavList.get(mainSecIdNavList.size() - 1).getDate();
                }

                List<IndexesTradeDate> indexesTradeDateDoList = TradeDateUtil.listInfoByDateInterval(theFirstNavDate, theLastNavDate);

                List<DateValue> validMainSecNavList = NavUtil.filterFrequencyNav(mainSecIdNavList, indexesTradeDateDoList, navFrequency);
                secAllNavMap.put(mainSecId, validMainSecNavList);

                Frequency calcFrequency = null;
                Frequency defaultFrequency = secFreqMap.get(mainSecId);
                if (dataFrequency == Frequency.Default) {
                    calcFrequency = defaultFrequency;
                } else {
                    calcFrequency = dataFrequency;
                }

                List<Indicator> indicatorList = dataFrequencyIndicatorListMap.get(dataFrequency);
                List<Indicator> geoExtraIndicatorList = dataFrequencyGeoExtraIndicatorListMap.get(dataFrequency);

                List<IndicatorCalcPropertyDto> indicatorCalcPropertyDtoList = new ArrayList<>();
                for (DateIntervalDto dateIntervalDto : dateIntervalDtoList) {

                    rollFindStartDate(mainSecId, dateIntervalDto, indexesTradeDateDoList, allNavMap, defaultFrequency);

                    List<IndicatorCalcTimeRangeDto> indicatorCalcTimeRangeDtoList = DateIntervalUtil.getIndicatorCalcTimeRangeDto(dateIntervalDto, validMainSecNavList, indexesTradeDateDoList
                            , dateIntervalDto.getId());

                    for (IndicatorCalcTimeRangeDto indicatorCalcTimeRangeDto : indicatorCalcTimeRangeDtoList) {

                        Map<String, List<DateValue>> dateSecAllNavMap = MapUtil.newHashMap(3 + indexIdList.size());
                        dateSecAllNavMap.put(mainSecId, NavUtil.getNewNavList(secAllNavMap.get(mainSecId)));
                        dateSecAllNavMap.put(benchmarkId, NavUtil.getNewNavList(secAllNavMap.get(benchmarkId)));
                        for (String indexId : indexIdList) {
                            dateSecAllNavMap.put(indexId, NavUtil.getNewNavList(secAllNavMap.get(indexId)));
                        }
                        if (StringUtils.isNotEmpty(riskOfFreeId)) {
                            dateSecAllNavMap.put(riskOfFreeId, NavUtil.getNewNavList(secAllNavMap.get(riskOfFreeId)));
                        }

                        if (validMainSecNavList.size() > 0) {
                            IndicatorCalcReqPropertyDto indicatorCalcReq = prepareIndicatorCalcReq(mainSecId, benchmarkId, indexIdList, indicatorCalcTimeRangeDto,
                                    dateSecAllNavMap, riskOfFreeId, riskOfFreeValue, null, calcFrequency, defaultFrequency, indicatorList, geoExtraIndicatorList, ifAnnualize, calcIndexRetIndicatorValue,
                                    needRet, needExtraRet, false);
                            indicatorCalcReq.setLeastRetNum(leastRetNum);

                            IndicatorCalcPropertyDto indicatorCalcPropertyDto = CalcService.getInstance().indicatorCalc(indicatorCalcReq);

                            indicatorCalcPropertyDto.getIndicatorCalcReq().setAllNavMap(null);
                            indicatorCalcPropertyDto.setRiskFreeData(null);
                            indicatorCalcPropertyDto.getSecData().setOriNavList(null);
                            indicatorCalcPropertyDto.getSecData().setExtraNavList(null);
                            indicatorCalcPropertyDto.getSecData().setRetList(null);
                            indicatorCalcPropertyDto.getSecData().setExtraRetGeoList(null);
                            indicatorCalcPropertyDto.getBenchmarkData().setBenchmarkNavList(null);
                            indicatorCalcPropertyDto.getBenchmarkData().setOriBenchmarkNavList(null);
                            indicatorCalcPropertyDto.getBenchmarkData().setBenchmarkRetList(null);

                            indicatorCalcPropertyDtoList.add(indicatorCalcPropertyDto);
                        }
                    }
                }

                indicatorCalcPropertyDtoListMap.put(dataFrequency, indicatorCalcPropertyDtoList);
            }
            indicatorCalcPropertyDtoListMapMap.put(mainSecId, indicatorCalcPropertyDtoListMap);
        }

        return indicatorCalcPropertyDtoListMapMap;
    }

    public List<IndicatorCalcTimeRangeDto> getSecTimeRange(String mainSecId, Frequency frequency, Map<String, List<DateValue>> allNavMap) {
        List<IndicatorCalcTimeRangeDto> indicatorCalcTimeRangeDtoList = CollectionUtil.newArrayList();

        DateIntervalDto dateIntervalDto = new DateIntervalDto();
        dateIntervalDto.setDateIntervalType(DateIntervalType.DefaultInterval);

        List<DateValue> validNavList = allNavMap.get(mainSecId);

        if (CollectionUtil.isEmpty(validNavList)) {
            return indicatorCalcTimeRangeDtoList;
        }

        for (DateValue dateValue : validNavList) {
            dateValue.setIsvalid(1);
        }

        List<DateValue> mainSecIdNavList = allNavMap.get(mainSecId);

        String theFirstNavDate = "";
        String theLastNavDate = "";
        if (mainSecIdNavList != null && mainSecIdNavList.size() > 0) {
            theFirstNavDate = mainSecIdNavList.get(0).getDate();
            theLastNavDate = mainSecIdNavList.get(mainSecIdNavList.size() - 1).getDate();
        }

        List<IndexesTradeDate> indexesTradeDateList = TradeDateUtil.listInfoByDateInterval(theFirstNavDate, theLastNavDate);

        indicatorCalcTimeRangeDtoList = DateIntervalUtil.getIndicatorCalcTimeRangeDto(dateIntervalDto, allNavMap.get(mainSecId), indexesTradeDateList
                , dateIntervalDto.getId());

        for (IndicatorCalcTimeRangeDto indicatorCalcTimeRangeDto : indicatorCalcTimeRangeDtoList) {
            if (frequency == Frequency.Monthly) {
                List<String> thisMonthAfter5Days = TradeDateUtil.getThisMonthAfter5Days(indicatorCalcTimeRangeDto.getStartDate(), indicatorCalcTimeRangeDto.getEndDate());
                indicatorCalcTimeRangeDto.setMonthPreDays(thisMonthAfter5Days);
                List<String> thisMonthBefore6Days = TradeDateUtil.getThisMonthBefore6Days(indicatorCalcTimeRangeDto.getStartDate(), indicatorCalcTimeRangeDto.getEndDate());
                indicatorCalcTimeRangeDto.setMonthLaterDays(thisMonthBefore6Days);
            } else {
                List<String> preFourTradeDays = TradeDateUtil.getBefore4TradeDate(indicatorCalcTimeRangeDto.getStartDate());
                indicatorCalcTimeRangeDto.setPreDays(preFourTradeDays);
            }
            List<DateValue> secIdValidNav = IndicatorCalcUtil.getSecIdValidNav(mainSecId, indicatorCalcTimeRangeDto, allNavMap, false, frequency);

            if (secIdValidNav.size() == 0) {
                indicatorCalcTimeRangeDto.setStartDate(null);
            }
        }

        return indicatorCalcTimeRangeDtoList;
    }

    public Map<String, List<IndicatorCalcPropertyDto>> getMultiSecRetListNew(List<String> mainSecIdList, List<String> indexIdList, Frequency frequency,
                                                                             Frequency rollingFrequency, String startDate, String endDate,
                                                                             boolean ifExcessReturn, String benchmarkId, NavType navType,
                                                                             Map<String, List<DateValue>> allNavMap, Map<String, Frequency> secFreqMap) {
        return this.getMultiSecRetListNew(mainSecIdList, indexIdList, frequency, rollingFrequency, startDate, endDate, ifExcessReturn,
                benchmarkId, RaiseType.Both, Strategy.All, Visibility.Both, navType, allNavMap, secFreqMap);
    }

    public Map<String, List<IndicatorCalcPropertyDto>> getMultiSecRetListNew(List<String> mainSecIdList, List<String> indexIdList, Frequency frequency,
                                                                             Frequency rollingFrequency, String startDate, String endDate,
                                                                             boolean ifExcessReturn, String benchmarkId, RaiseType raiseType, IStrategy strategy,
                                                                             Visibility visibility, NavType navType,
                                                                             Map<String, List<DateValue>> allNavMap, Map<String, Frequency> secFreqMap) {
        List<DateIntervalDto> dateIntervalDtoList = ListUtil.list(false);
        DateIntervalDto defaultIntervalDto = new DateIntervalDto();
        defaultIntervalDto.setStartDate(startDate);
        defaultIntervalDto.setEndDate(endDate);
        defaultIntervalDto.setDateIntervalType(DateIntervalType.DefaultRolling);
        defaultIntervalDto.setFrequency(rollingFrequency);
        dateIntervalDtoList.add(defaultIntervalDto);

        List<Indicator> indicatorList = ListUtil.list(false);
        List<Indicator> geoExtraIndicatorList = ListUtil.list(false);

        indicatorList.add(Indicator.IntervalReturn);
        if (ifExcessReturn) {
            geoExtraIndicatorList.add(Indicator.IntervalReturn);
        }

        Map<String, String> secBenchmarkIdMap = MapUtil.newHashMap(mainSecIdList.size());
        for (String mainSecId : mainSecIdList) {
            secBenchmarkIdMap.put(mainSecId, benchmarkId);
        }

        Map<String, List<DateIntervalDto>> secDateIntervalDtoListMap = MapUtil.newHashMap(mainSecIdList.size());
        for (String mainSecId : mainSecIdList) {
            secDateIntervalDtoListMap.put(mainSecId, dateIntervalDtoList);
        }

        CalcMultipleSecMultipleTimeRangeIndicatorReq calcMultipleSecMultipleTimeRangeIndicatorReq = new CalcMultipleSecMultipleTimeRangeIndicatorReq();
        calcMultipleSecMultipleTimeRangeIndicatorReq.setMainSecIdList(mainSecIdList);
        calcMultipleSecMultipleTimeRangeIndicatorReq.setSecBenchmarkIdMap(secBenchmarkIdMap);
        calcMultipleSecMultipleTimeRangeIndicatorReq.setIndexIdList(indexIdList);
        calcMultipleSecMultipleTimeRangeIndicatorReq.setRaiseType(raiseType);
        calcMultipleSecMultipleTimeRangeIndicatorReq.setStrategy(strategy);
        calcMultipleSecMultipleTimeRangeIndicatorReq.setVisibility(visibility);
        calcMultipleSecMultipleTimeRangeIndicatorReq.setSecDateIntervalDtoListMap(secDateIntervalDtoListMap);
        calcMultipleSecMultipleTimeRangeIndicatorReq.setDataFrequency(frequency);
        calcMultipleSecMultipleTimeRangeIndicatorReq.setNavType(navType);
        calcMultipleSecMultipleTimeRangeIndicatorReq.setIndicatorList(indicatorList);
        calcMultipleSecMultipleTimeRangeIndicatorReq.setGeoExtraindicatorList(geoExtraIndicatorList);
        calcMultipleSecMultipleTimeRangeIndicatorReq.setRiskOfFreeId(null);
        calcMultipleSecMultipleTimeRangeIndicatorReq.setRiskOfFreeValue(null);
        calcMultipleSecMultipleTimeRangeIndicatorReq.setIfAnnualize(true);
        calcMultipleSecMultipleTimeRangeIndicatorReq.setCalcIndexRetIndicatorValue(true);
        calcMultipleSecMultipleTimeRangeIndicatorReq.setNeedRet(false);
        calcMultipleSecMultipleTimeRangeIndicatorReq.setNeedExtraRet(false);
        calcMultipleSecMultipleTimeRangeIndicatorReq.setIfConvertPerformanceConsistencyWord(true);

        return calcMultipleSecMultipleTimeRangeIndicator(calcMultipleSecMultipleTimeRangeIndicatorReq, allNavMap, secFreqMap);
    }

    private IndicatorCalcReqPropertyDto prepareTrendReq(String mainSecId, String benchmarkId, List<String> indexList, IndicatorCalcTimeRangeDto indicatorCalcTimeRangeDto,
                                                        Map<String, List<DateValue>> allNavMap, Frequency frequency, Frequency defaultFrequency,
                                                        List<TrendType> trendTypeV2List, BigDecimal fixedIncome, BigDecimal initValue) {

        IndicatorCalcReqPropertyDto indicatorCalcReq = new IndicatorCalcReqPropertyDto();
        indicatorCalcReq.setSecId(mainSecId);
        indicatorCalcReq.setBenchmarkId(benchmarkId);
        indicatorCalcReq.setIndexIdList(indexList);
        indicatorCalcReq.setAllNavMap(allNavMap);
        indicatorCalcReq.setTrendTypeList(trendTypeV2List);
        indicatorCalcReq.setFixedIncome(fixedIncome);
        indicatorCalcReq.setFrequency(frequency);
        indicatorCalcReq.setInitValue(initValue);
        indicatorCalcReq.setDefaultFrequency(defaultFrequency);

        if (frequency == Frequency.Monthly) {
            List<String> thisMonthAfter5Days = TradeDateUtil.getThisMonthAfter5Days(indicatorCalcTimeRangeDto.getStartDate(), indicatorCalcTimeRangeDto.getEndDate());
            indicatorCalcTimeRangeDto.setMonthPreDays(thisMonthAfter5Days);
            List<String> thisMonthBefore6Days = TradeDateUtil.getThisMonthBefore6Days(indicatorCalcTimeRangeDto.getStartDate(), indicatorCalcTimeRangeDto.getEndDate());
            indicatorCalcTimeRangeDto.setMonthLaterDays(thisMonthBefore6Days);
        } else {
            List<String> preFourTradeDays = TradeDateUtil.getBefore4TradeDate(indicatorCalcTimeRangeDto.getStartDate());
            indicatorCalcTimeRangeDto.setPreDays(preFourTradeDays);
        }

        indicatorCalcReq.setIndicatorCalcTimeRangeDto(indicatorCalcTimeRangeDto);
        return indicatorCalcReq;
    }

    private IndicatorCalcReqPropertyDto prepareIndicatorCalcReq(String mainSecId, String benchmarkId, List<String> indexList,
                                                                IndicatorCalcTimeRangeDto indicatorCalcTimeRangeDto,
                                                                Map<String, List<DateValue>> allNavMap,
                                                                String riskOfFreeId, Double riskOfFreeValue, Map<String, CompoundRet> stgDateRetMap, Frequency frequency, Frequency defaultFrequency,
                                                                List<Indicator> indicatorList, List<Indicator> geoExtraindicatorList, boolean ifAnnualize, boolean calcIndexRetIndicatorValue,
                                                                boolean needRet, boolean needExtraRet, boolean ifConvertPerformanceConsistencyWord) {
        IndicatorCalcReqPropertyDto indicatorCalcReq = new IndicatorCalcReqPropertyDto();
        indicatorCalcReq.setSecId(mainSecId);
        if (indexList != null) {
            indicatorCalcReq.setIndexIdList(indexList);
        }
        indicatorCalcReq.setBenchmarkId(benchmarkId);
        indicatorCalcReq.setAllNavMap(allNavMap);

        if (defaultFrequency == Frequency.Monthly) {
            List<String> thisMonthAfter5Days = TradeDateUtil.getThisMonthAfter5Days(indicatorCalcTimeRangeDto.getStartDate(), indicatorCalcTimeRangeDto.getEndDate());
            indicatorCalcTimeRangeDto.setMonthPreDays(thisMonthAfter5Days);
            List<String> thisMonthBefore6Days = TradeDateUtil.getThisMonthBefore6Days(indicatorCalcTimeRangeDto.getStartDate(), indicatorCalcTimeRangeDto.getEndDate());
            indicatorCalcTimeRangeDto.setMonthLaterDays(thisMonthBefore6Days);
        } else {
            List<String> preFourTradeDays = TradeDateUtil.getBefore4TradeDate(indicatorCalcTimeRangeDto.getStartDate());
            indicatorCalcTimeRangeDto.setPreDays(preFourTradeDays);
        }

        indicatorCalcReq.setIndicatorCalcTimeRangeDto(indicatorCalcTimeRangeDto);

        indicatorCalcReq.setStrategyDateRetMap(stgDateRetMap);
        indicatorCalcReq.setRiskOfFreeId(riskOfFreeId);
        indicatorCalcReq.setRiskOfFreeValue(riskOfFreeValue);
        indicatorCalcReq.setIndicatorList(indicatorList);
        indicatorCalcReq.setGeoIndicatorList(geoExtraindicatorList);

        indicatorCalcReq.setFrequency(frequency);
        indicatorCalcReq.setDefaultFrequency(defaultFrequency);

        indicatorCalcReq.setIfAnnualize(ifAnnualize);

        indicatorCalcReq.setNeedRet(needRet);

        indicatorCalcReq.setNeedExtraGeoRet(needExtraRet);

        indicatorCalcReq.setCalcIndexRetIndicatorValue(calcIndexRetIndicatorValue);

        indicatorCalcReq.setIfConvertPerformanceConsistencyWord(ifConvertPerformanceConsistencyWord);

        return indicatorCalcReq;
    }

    public void rollFindStartDate(String mainSecId, DateIntervalDto dateIntervalDto, List<IndexesTradeDate> indexesTradeDateList, Map<String, List<DateValue>> allNavMap, Frequency defaultFrequency) {
        if (dateIntervalDto.getDateIntervalType() == DateIntervalType.DefaultRolling || dateIntervalDto.getDateIntervalType() == DateIntervalType.CustomRolling) {
            DateIntervalDto defaultIntervalDateIntervalDto = new DateIntervalDto();
            defaultIntervalDateIntervalDto.setStartDate(dateIntervalDto.getStartDate());
            defaultIntervalDateIntervalDto.setEndDate(dateIntervalDto.getEndDate());
            defaultIntervalDateIntervalDto.setDateIntervalType(DateIntervalType.CustomInterval);
            defaultIntervalDateIntervalDto.setTimeRange(dateIntervalDto.getTimeRange());

            List<DateValue> newNavList = NavUtil.getNewNavList(allNavMap.get(mainSecId));
            List<IndicatorCalcTimeRangeDto> indicatorCalcTimeRangeDtoList = DateIntervalUtil.getIndicatorCalcTimeRangeDto(defaultIntervalDateIntervalDto, newNavList, indexesTradeDateList
                    , dateIntervalDto.getId());

            if (indicatorCalcTimeRangeDtoList != null && indicatorCalcTimeRangeDtoList.size() > 0) {
                IndicatorCalcTimeRangeDto indicatorCalcTimeRangeDto = indicatorCalcTimeRangeDtoList.get(0);
                if (defaultFrequency == Frequency.Monthly) {
                    List<String> thisMonthAfter5Days = TradeDateUtil.getThisMonthAfter5Days(indicatorCalcTimeRangeDto.getStartDate(), indicatorCalcTimeRangeDto.getEndDate());
                    indicatorCalcTimeRangeDto.setMonthPreDays(thisMonthAfter5Days);
                    List<String> thisMonthBefore6Days = TradeDateUtil.getThisMonthBefore6Days(indicatorCalcTimeRangeDto.getStartDate(), indicatorCalcTimeRangeDto.getEndDate());
                    indicatorCalcTimeRangeDto.setMonthLaterDays(thisMonthBefore6Days);
                } else {
                    List<String> preFourTradeDays = TradeDateUtil.getBefore4TradeDate(indicatorCalcTimeRangeDto.getStartDate());
                    indicatorCalcTimeRangeDto.setPreDays(preFourTradeDays);
                }
                Map<String, List<DateValue>> defaultIntervalNavMap = MapUtil.newHashMap(4);
                defaultIntervalNavMap.put(mainSecId, newNavList);
                List<DateValue> secIdValidNav = IndicatorCalcUtil.getSecIdValidNav(mainSecId, indicatorCalcTimeRangeDto, defaultIntervalNavMap, false, defaultFrequency);
                if (secIdValidNav != null && secIdValidNav.size() > 0) {
                    dateIntervalDto.setStartDate(secIdValidNav.get(0).getDate());
                }
            }
        }
    }

    private static class IndicatorServiceBuilder {
        public static final IndicatorService INSTANCE = new IndicatorService();
    }
}

